all_close_objects_set = [
    "cabinet",
    "drawer"
]

all_open_objects_set = [
    "cabinet",
    "drawer"
]

cabinet_objects_set = [
    "cabinet"
]

drawer_objects_set = [
    "drawer"
]

lite_pick_objects_set = [
    "bottled_drink_2",
    "cup_2",
    "egg_0",
    "hot_dog_0",
    "ketchup_2",
    "kiwi_0",
    "lemon_0",
    "mango_5",
    "orange_2",
    "potato_2",
    "tangerine_3",
    "tomato_1",
    "wine_4",
    "apple_3"
]

diverse_pick_objects_set = [
    "apple_6",
    "tomato_7",
    "can_16",
    "onion_1",
    "bottled_drink_0",
    "kettle_14",
    "spoon_1",
    "mango_19",
    "potato_2",
    "lime_1",
    "can_4",
    "apple_10",
    "water_bottle_4",
    "egg_1",
    "potato_5",
    "egg_4",
    "cupcake_17",
    "hot_dog_3",
    "can_6",
    "mango_11",
    "mushroom_3",
    "bagel_0",
    "candle_2",
    "rolling_pin_0",
    "croissant_2",
    "mug_7",
    "pear_2",
    "kettle_26",
    "bagel_6",
    "bar_3",
    "milk_5",
    "spoon_16",
    "squash_7",
    "sponge_1",
    "eggplant_0",
    "cupcake_3",
    "cereal_12",
    "bowl_19",
    "bread_8",
    "mug_18",
    "donut_0",
    "kettle_11",
    "pan_13",
    "knife_1",
    "bread_13",
    "plate_13",
    "mug_0",
    "bowl_3",
]

plausible_categories = [
    "alcohol",
    "apple",
    "avocado",
    "bagel",
    "baggedfood",
    "baguette",
    "banana",
    "bar",
    "barsoap",
    "beer",
    "bellpepper",
    "bottleddrink",
    "bottledwater",
    "bowl",
    "boxeddrink",
    "boxedfood",
    "bread",
    "broccoli",
    "cake",
    "can",
    "candle",
    "cannedfood",
    "carrot",
    "cereal",
    "cheese",
    "chips",
    "coffeecup",
    "condiment",
    "corn",
    "croissant",
    "cup",
    "cupcake",
    "cucumber",
    "donut",
    "egg",
    "eggplant",
    "fish",
    "fork",
    "garlic",
    "hotdog",
    "jam",
    "jug",
    "ketchup",
    "kettle",
    "kiwi",
    "knife",
    "ladle",
    "lemon",
    "lime",
    "mango",
    "milk",
    "mug",
    "mushroom",
    "onion",
    "orange",
    "pan",
    "peach",
    "pear",
    "potato",
    "rollingpin",
    "scissors",
    "shaker",
    "soapdispenser",
    "spatula",
    "sponge",
    "spoon",
    "spray",
    "squash",
    "steak",
    "sweetpotato",
    "tangerine",
    "teapot",
    "tomato",
    "waffle",
    "waterbottle",
    "wine",
    "yogurt"
]


all_categories_names = [
   "alcohol",
    "apple",
    "avocado",
    "bagel",
    "baggedfood",
    "baguette",
    "banana",
    "bar",
    "barsoap",
    "beer",
    "bellpepper",
    "bottleddrink",
    "bottledwater",
    "bowl",
    "boxeddrink",
    "boxedfood",
    "bread",
    "broccoli",
    "cake",
    "can",
    "candle",
    "cannedfood",
    "carrot",
    "cereal",
    "cheese",
    "chips",
    "coffeecup",
    "condiment",
    "corn",
    "croissant",
    "cup",
    "cucumber", 
    "cupcake",
    "cucumber",
    "donut",
    "egg",
    "eggplant",
    "fish",
    "fork",
    "garlic",
    "hotdog",
    "jam",
    "jug",
    "ketchup",
    "kettle",
    "kiwi",
    "knife",
    "ladle",
    "lemon",
    "lime",
    "mango",
    "milk",
    "mug",
    "mushroom",
    "onion",
    "orange",
    "pan",
    "peach",
    "pear",
    "potato",
    "rollingpin",
    "scissors",
    "shaker",
    "soapdispenser",
    "spatula",
    "sponge",
    "spoon",
    "spray",
    "squash",
    "steak",
    "sweetpotato",
    "tangerine",
    "teapot",
    "tomato",
    "waffle",
    "waterbottle",
    "wine",
    "yogurt"
]

full_eval_objects_set = [
    "alcohol_0",
    "alcohol_1",
    "alcohol_2",
    "alcohol_4",
    "alcohol_5",
    "alcohol_7",
    "alcohol_9",
    "alcohol_10",
    "alcohol_11",
    "alcohol_12",
    "apple_0",
    "apple_1",
    "apple_2",
    "apple_3",
    "apple_4",
    "apple_5",
    "apple_6",
    "apple_7",
    "apple_8",
    "apple_9",
    "avocado_0",
    "avocado_1",
    "avocado_2",
    "avocado_4",
    "avocado_5",
    "avocado_6",
    "avocado_8",
    "bagel_0",
    "bagel_1",
    "bagel_2",
    "bagel_3",
    "bagel_4",
    "bagel_5",
    "bagel_6",
    "bagel_7",
    "bagel_9",
    "bagged_food_0",
    "bagged_food_3",
    "bagged_food_4",
    "bagged_food_5",
    "bagged_food_7",
    "bagged_food_8",
    "bagged_food_9",
    "bagged_food_11",
    "bagged_food_14",
    "baguette_0",
    "baguette_1",
    "baguette_2",
    "baguette_4",
    "baguette_5",
    "baguette_6",
    "banana_1",
    "banana_2",
    "banana_3",
    "banana_4",
    "banana_5",
    "banana_6",
    "banana_7",
    "banana_8",
    "banana_9",
    "banana_10",
    "bar_0",
    "bar_2",
    "bar_3",
    "bar_4",
    "bar_6",
    "bar_7",
    "bar_8",
    "bar_9",
    "bar_10",
    "bar_11",
    "bar_soap_0",
    "bar_soap_1",
    "bar_soap_3",
    "bar_soap_4",
    "beer_0",
    "beer_1",
    "beer_2",
    "beer_3",
    "beer_5",
    "beer_6",
    "beer_7",
    "beer_8",
    "beer_9",
    "beer_10",
    "bell_pepper_0",
    "bell_pepper_1",
    "bell_pepper_2",
    "bell_pepper_3",
    "bottled_drink_0",
    "bottled_drink_1",
    "bottled_drink_2",
    "bottled_drink_4",
    "bottled_drink_5",
    "bottled_water_1",
    "bottled_water_2",
    "bottled_water_3",
    "bottled_water_4",
    "bottled_water_6",
    "bottled_water_7",
    "bottled_water_8",
    "bottled_water_10",
    "bottled_water_11",
    "bottled_water_12",
    "bowl_0",
    "bowl_1",
    "bowl_2",
    "bowl_3",
    "bowl_4",
    "bowl_5",
    "bowl_6",
    "bowl_7",
    "bowl_8",
    "bowl_9",
    "boxed_drink_0",
    "boxed_drink_1",
    "boxed_drink_2",
    "boxed_drink_3",
    "boxed_drink_4",
    "boxed_drink_5",
    "boxed_drink_7",
    "boxed_food_0",
    "boxed_food_1",
    "boxed_food_2",
    "boxed_food_4",
    "boxed_food_6",
    "boxed_food_7",
    "boxed_food_8",
    "boxed_food_9",
    "boxed_food_10",
    "boxed_food_11",
    "bread_0",
    "bread_1",
    "bread_2",
    "bread_3",
    "bread_4",
    "bread_5",
    "bread_6",
    "bread_7",
    "bread_8",
    "bread_9",
    "broccoli_0",
    "broccoli_1",
    "broccoli_3",
    "cake_0",
    "cake_1",
    "cake_3",
    "cake_4",
    "cake_5",
    "cake_6",
    "cake_7",
    "cake_8",
    "cake_9",
    "can_0",
    "can_2",
    "can_3",
    "can_4",
    "can_6",
    "can_8",
    "can_9",
    "can_11",
    "can_12",
    "can_13",
    "candle_0",
    "candle_1",
    "candle_2",
    "candle_3",
    "candle_4",
    "candle_5",
    "candle_6",
    "candle_7",
    "candle_8",
    "candle_9",
    "canned_food_0",
    "canned_food_1",
    "canned_food_3",
    "canned_food_4",
    "canned_food_5",
    "canned_food_6",
    "canned_food_8",
    "canned_food_9",
    "canned_food_10",
    "canned_food_11",
    "carrot_0",
    "carrot_1",
    "carrot_2",
    "carrot_3",
    "carrot_4",
    "carrot_6",
    "carrot_7",
    "cereal_0",
    "cereal_1",
    "cereal_2",
    "cereal_3",
    "cereal_4",
    "cereal_5",
    "cereal_6",
    "cereal_7",
    "cereal_8",
    "cereal_9",
    "cheese_0",
    "cheese_1",
    "cheese_2",
    "cheese_3",
    "cheese_4",
    "cheese_5",
    "cheese_6",
    "cheese_7",
    "cheese_8",
    "chips_0",
    "chips_1",
    "chips_2",
    "chips_3",
    "chips_4",
    "chips_5",
    "chips_6",
    "chips_7",
    "chips_8",
    "chips_9",
    "coffee_cup_1",
    "coffee_cup_2",
    "coffee_cup_3",
    "coffee_cup_4",
    "coffee_cup_6",
    "coffee_cup_7",
    "coffee_cup_8",
    "coffee_cup_9",
    "coffee_cup_10",
    "coffee_cup_11",
    "condiment_0",
    "condiment_2",
    "condiment_3",
    "corn_0",
    "corn_1",
    "corn_3",
    "corn_4",
    "croissant_0",
    "croissant_1",
    "croissant_2",
    "croissant_3",
    "croissant_4",
    "croissant_5",
    "croissant_6",
    "croissant_8",
    "croissant_9",
    "croissant_10",
    "cucumber_0",
    "cucumber_1",
    "cucumber_2",
    "cucumber_3",
    "cucumber_4",
    "cucumber_5",
    "cup_2",
    "cup_3",
    "cup_4",
    "cup_5",
    "cupcake_2",
    "cupcake_3",
    "cupcake_4",
    "cupcake_5",
    "cupcake_6",
    "cupcake_7",
    "cupcake_9",
    "cupcake_11",
    "cupcake_12",
    "cupcake_14",
    "cutting_board_0",
    "cutting_board_1",
    "cutting_board_2",
    "cutting_board_4",
    "cutting_board_5",
    "cutting_board_7",
    "cutting_board_9",
    "cutting_board_11",
    "cutting_board_12",
    "cutting_board_13",
    "donut_0",
    "donut_1",
    "donut_2",
    "donut_3",
    "donut_4",
    "donut_5",
    "donut_6",
    "donut_7",
    "donut_8",
    "donut_9",
    "egg_0",
    "egg_1",
    "egg_3",
    "egg_4",
    "egg_5",
    "egg_6",
    "egg_7",
    "egg_8",
    "egg_9",
    "egg_10",
    "eggplant_0",
    "eggplant_1",
    "eggplant_2",
    "eggplant_3",
    "eggplant_4",
    "fish_0",
    "fish_1",
    "fish_2",
    "fish_3",
    "fish_4",
    "fish_5",
    "fish_6",
    "fish_7",
    "fish_8",
    "fish_9",
    "fork_0",
    "fork_1",
    "fork_2",
    "fork_3",
    "fork_4",
    "fork_5",
    "fork_6",
    "fork_7",
    "fork_8",
    "fork_9",
    "garlic_0",
    "garlic_1",
    "garlic_2",
    "garlic_4",
    "garlic_5",
    "hot_dog_0",
    "hot_dog_1",
    "hot_dog_2",
    "hot_dog_3",
    "hot_dog_4",
    "hot_dog_5",
    "hot_dog_6",
    "jam_0",
    "jam_1",
    "jam_2",
    "jam_3",
    "jam_4",
    "jam_5",
    "jam_6",
    "jug_0",
    "jug_1",
    "jug_2",
    "jug_3",
    "jug_4",
    "jug_5",
    "jug_6",
    "jug_7",
    "jug_8",
    "ketchup_0",
    "ketchup_1",
    "ketchup_2",
    "ketchup_3",
    "ketchup_4",
    "ketchup_7",
    "ketchup_8",
    "ketchup_9",
    "ketchup_10",
    "ketchup_11",
    "kettle_0",
    "kettle_1",
    "kettle_2",
    "kettle_3",
    "kettle_4",
    "kettle_5",
    "kettle_6",
    "kettle_7",
    "kettle_8",
    "kettle_9",
    "kiwi_0",
    "kiwi_1",
    "kiwi_2",
    "kiwi_4",
    "kiwi_5",
    "kiwi_6",
    "kiwi_7",
    "knife_0",
    "knife_1",
    "knife_2",
    "knife_3",
    "knife_4",
    "knife_5",
    "knife_6",
    "knife_7",
    "knife_8",
    "knife_9",
    "ladle_0",
    "ladle_1",
    "ladle_2",
    "ladle_3",
    "ladle_4",
    "lemon_0",
    "lemon_1",
    "lemon_2",
    "lemon_3",
    "lemon_4",
    "lemon_5",
    "lemon_6",
    "lemon_7",
    "lemon_8",
    "lemon_9",
    "lime_0",
    "lime_1",
    "lime_2",
    "lime_3",
    "mango_1",
    "mango_4",
    "mango_5",
    "mango_7",
    "mango_8",
    "mango_9",
    "mango_10",
    "mango_11",
    "mango_12",
    "mango_13",
    "milk_0",
    "milk_1",
    "milk_3",
    "milk_4",
    "milk_5",
    "milk_7",
    "milk_9",
    "milk_10",
    "mug_0",
    "mug_1",
    "mug_2",
    "mug_3",
    "mug_4",
    "mug_5",
    "mug_7",
    "mug_8",
    "mug_9",
    "mug_10",
    "mushroom_0",
    "mushroom_1",
    "mushroom_2",
    "mushroom_3",
    "mushroom_4",
    "mushroom_5",
    "mushroom_6",
    "mushroom_7",
    "mushroom_8",
    "mushroom_9",
    "onion_0",
    "onion_1",
    "onion_2",
    "onion_3",
    "onion_4",
    "onion_6",
    "onion_7",
    "onion_8",
    "onion_9",
    "onion_10",
    "orange_0",
    "orange_1",
    "orange_2",
    "orange_3",
    "orange_4",
    "orange_5",
    "orange_8",
    "orange_9",
    "orange_10",
    "orange_11",
    "pan_0",
    "pan_1",
    "pan_2",
    "pan_3",
    "pan_4",
    "pan_5",
    "pan_6",
    "pan_7",
    "pan_8",
    "pan_9",
    "peach_0",
    "peach_1",
    "peach_2",
    "peach_3",
    "peach_4",
    "peach_5",
    "peach_6",
    "pear_0",
    "pear_1",
    "pear_2",
    "pear_3",
    "pear_5",
    "pear_6",
    "pear_8",
    "pear_10",
    "pear_11",
    "pear_12",
    "plate_0",
    "plate_1",
    "plate_2",
    "plate_3",
    "plate_4",
    "plate_5",
    "plate_7",
    "plate_8",
    "plate_9",
    "plate_10",
    "potato_0",
    "potato_1",
    "potato_2",
    "potato_3",
    "potato_4",
    "potato_5",
    "potato_6",
    "potato_7",
    "potato_8",
    "potato_9",
    "rolling_pin_0",
    "rolling_pin_1",
    "rolling_pin_2",
    "rolling_pin_3",
    "rolling_pin_4",
    "rolling_pin_5",
    "scissors_0",
    "scissors_1",
    "scissors_3",
    "scissors_4",
    "scissors_6",
    "scissors_8",
    "scissors_9",
    "scissors_10",
    "scissors_11",
    "scissors_12",
    "shaker_0",
    "shaker_1",
    "shaker_2",
    "shaker_3",
    "soap_dispenser_0",
    "soap_dispenser_1",
    "soap_dispenser_2",
    "soap_dispenser_3",
    "soap_dispenser_4",
    "soap_dispenser_6",
    "spatula_0",
    "spatula_1",
    "spatula_2",
    "spatula_3",
    "spatula_4",
    "spatula_5",
    "spatula_6",
    "spatula_7",
    "spatula_8",
    "spatula_9",
    "sponge_0",
    "sponge_1",
    "sponge_2",
    "sponge_3",
    "sponge_4",
    "sponge_5",
    "sponge_6",
    "sponge_8",
    "spoon_0",
    "spoon_1",
    "spoon_2",
    "spoon_3",
    "spoon_4",
    "spoon_5",
    "spoon_6",
    "spoon_7",
    "spoon_8",
    "spoon_9",
    "spray_0",
    "spray_1",
    "spray_2",
    "squash_0",
    "squash_1",
    "squash_3",
    "squash_5",
    "squash_7",
    "squash_8",
    "squash_13",
    "squash_15",
    "squash_16",
    "steak_2",
    "steak_3",
    "steak_4",
    "steak_5",
    "steak_6",
    "steak_8",
    "steak_9",
    "steak_10",
    "steak_11",
    "steak_14",
    "sweet_potato_1",
    "sweet_potato_3",
    "tangerine_0",
    "tangerine_1",
    "tangerine_2",
    "tangerine_3",
    "tangerine_4",
    "tangerine_5",
    "tangerine_6",
    "teapot_0",
    "teapot_3",
    "teapot_4",
    "teapot_5",
    "teapot_6",
    "teapot_7",
    "teapot_10",
    "teapot_11",
    "teapot_13",
    "teapot_14",
    "tomato_0",
    "tomato_1",
    "tomato_2",
    "tomato_3",
    "tomato_5",
    "tomato_6",
    "tomato_7",
    "tomato_8",
    "waffle_0",
    "waffle_1",
    "waffle_3",
    "waffle_4",
    "water_bottle_0",
    "water_bottle_1",
    "water_bottle_2",
    "water_bottle_3",
    "water_bottle_4",
    "water_bottle_5",
    "water_bottle_6",
    "water_bottle_7",
    "water_bottle_8",
    "water_bottle_9",
    "wine_1",
    "wine_2",
    "wine_3",
    "wine_4",
    "wine_5",
    "wine_6",
    "wine_8",
    "wine_9",
    "wine_10",
    "wine_11",
    "yogurt_1",
    "yogurt_2",
    "yogurt_3",
]

all_pick_objects_set = [
    "kettle_1",
    "spoon_14",
    "egg_5",
    "kettle_13",
    "cheese_8",
    "lemon_16",
    "water_bottle_17",
    "banana_23",
    "bagel_6",
    "spatula_6",
    "water_bottle_5",
    "beer_8",
    "condiment_0",
    "bread_1",
    "bagel_2",
    "mango_5",
    "can_11",
    "cutting_board_12",
    "fish_0",
    "cheese_6",
    "spoon_17",
    "jug_7",
    "peach_1",
    "cheese_2",
    "avocado_5",
    "bowl_19",
    "wine_2",
    "ketchup_9",
    "tray_1",
    "mango_4",
    "water_bottle_23",
    "cake_4",
    "cucumber_5",
    "fish_7",
    "apple_8",
    "bottled_water_6",
    "donut_13",
    "mushroom_13",
    "fish_3",
    "plate_15",
    "potato_7",
    "plate_13",
    "banana_21",
    "sponge_6",
    "wine_11",
    "bagged_food_8",
    "bowl_17",
    "cereal_5",
    "kiwi_0",
    "boxed_drink_7",
    "donut_1",
    "cutting_board_4",
    "candle_12",
    "chocolate_6",
    "sponge_4",
    "chips_1",
    "cereal_7",
    "lemon_4",
    "coffee_cup_16",
    "broccoli_0",
    "cereal_11",
    "sponge_2",
    "bell_pepper_0",
    "cupcake_17",
    "bar_soap_3",
    "steak_4",
    "boxed_drink_3",
    "sweet_potato_1",
    "plate_3",
    "can_3",
    "sponge_8",
    "milk_3",
    "kettle_19",
    "scissors_10",
    "boxed_food_1",
    "cake_3",
    "mug_18",
    "plate_10",
    "mushroom_7",
    "water_bottle_6",
    "hot_dog_0",
    "squash_16",
    "peach_0",
    "squash_3",
    "jug_4",
    "mango_16",
    "mushroom_12",
    "bread_6",
    "fork_0",
    "bar_3",
    "kettle_0",
    "mango_9",
    "chips_2",
    "wine_6",
    "coffee_cup_17",
    "coffee_cup_21",
    "potato_1",
    "milk_5",
    "wine_3",
    "wine_1",
    "kettle_23",
    "orange_13",
    "orange_0",
    "cup_3",
    "croissant_9",
    "apple_4",
    "spatula_1",
    "mushroom_0",
    "pear_16",
    "water_bottle_1",
    "carrot_4",
    "cheese_4",
    "kettle_9",
    "beer_16",
    "pan_9",
    "cupcake_18",
    "waffle_1",
    "spoon_6",
    "plate_11",
    "milk_9",
    "steak_2",
    "water_bottle_13",
    "orange_3",
    "fork_5",
    "fork_11",
    "cutting_board_16",
    "spray_2",
    "plate_1",
    "lime_2",
    "bottled_drink_0",
    "bowl_12",
    "kettle_17",
    "kettle_4",
    "peach_3",
    "hot_dog_3",
    "mushroom_3",
    "scissors_8",
    "beer_12",
    "alcohol_13",
    "spatula_12",
    "tray_9",
    "bread_21",
    "hot_dog_1",
    "croissant_1",
    "bread_4",
    "banana_5",
    "boxed_drink_2",
    "fork_6",
    "cupcake_9",
    "mushroom_10",
    "potato_13",
    "candle_1",
    "teapot_4",
    "mug_8",
    "beer_0",
    "pan_21",
    "shaker_0",
    "pan_19",
    "peach_5",
    "apple_9",
    "canned_food_8",
    "canned_food_10",
    "ketchup_2",
    "water_bottle_8",
    "yogurt_1",
    "spoon_8",
    "pear_6",
    "cheese_1",
    "spray_0",
    "knife_17",
    "fork_1",
    "sweet_potato_3",
    "scissors_4",
    "potato_12",
    "broccoli_3",
    "plate_14",
    "banana_16",
    "jug_3",
    "bottled_water_13",
    "bell_pepper_2",
    "avocado_2",
    "pan_23",
    "squash_8",
    "coffee_cup_8",
    "banana_6",
    "bowl_16",
    "bread_18",
    "coffee_cup_12",
    "egg_12",
    "water_bottle_18",
    "bread_13",
    "soap_dispenser_2",
    "cake_0",
    "chocolate_11",
    "lemon_8",
    "knife_15",
    "corn_0",
    "teapot_7",
    "water_bottle_7",
    "canned_food_9",
    "cucumber_1",
    "water_bottle_10",
    "garlic_1",
    "jam_2",
    "knife_2",
    "bowl_1",
    "lime_0",
    "donut_9",
    "lemon_2",
    "beer_9",
    "bowl_18",
    "bowl_9",
    "chocolate_3",
    "tomato_8",
    "spatula_13",
    "pan_18",
    "cake_1",
    "ladle_2",
    "tomato_6",
    "teapot_15",
    "bowl_6",
    "spoon_5",
    "wine_12",
    "squash_15",
    "spoon_18",
    "bread_14",
    "lemon_9",
    "water_bottle_12",
    "pan_8",
    "squash_0",
    "ketchup_0",
    "can_13",
    "waffle_0",
    "jam_4",
    "bar_2",
    "tangerine_3",
    "mushroom_11",
    "boxed_drink_0",
    "carrot_1",
    "water_bottle_19",
    "banana_13",
    "ladle_0",
    "beer_7",
    "chips_8",
    "bottled_water_2",
    "chocolate_12",
    "alcohol_7",
    "egg_13",
    "potato_2",
    "bottled_drink_2",
    "candle_13",
    "kiwi_7",
    "can_0",
    "apple_13",
    "cereal_8",
    "pear_18",
    "tangerine_4",
    "plate_12",
    "knife_1",
    "pear_8",
    "bowl_3",
    "onion_0",
    "kettle_3",
    "scissors_0",
    "jug_1",
    "chocolate_13",
    "banana_18",
    "donut_14",
    "croissant_0",
    "cup_5",
    "onion_9",
    "knife_3",
    "garlic_5",
    "bowl_13",
    "chocolate_8",
    "jam_6",
    "baguette_0",
    "cupcake_6",
    "carrot_3",
    "knife_5",
    "tray_11",
    "kettle_8",
    "croissant_4",
    "knife_19",
    "spatula_4",
    "scissors_6",
    "coffee_cup_3",
    "egg_3",
    "plate_17",
    "can_2",
    "canned_food_6",
    "water_bottle_20",
    "bar_10",
    "jug_8",
    "chocolate_7",
    "pear_12",
    "knife_27",
    "kettle_24",
    "hot_dog_4",
    "peach_2",
    "lime_3",
    "mango_12",
    "mushroom_4",
    "bottled_drink_1",
    "donut_7",
    "cutting_board_13",
    "can_17",
    "bottled_water_12",
    "yogurt_3",
    "chocolate_1",
    "cutting_board_5",
    "eggplant_4",
    "scissors_13",
    "spatula_5",
    "bar_4",
    "bowl_11",
    "ketchup_8",
    "knife_6",
    "plate_0",
    "candle_3",
    "fish_1",
    "lemon_10",
    "waffle_4",
    "water_bottle_21",
    "potato_3",
    "candle_10",
    "spoon_22",
    "spatula_9",
    "fork_4",
    "bagel_9",
    "spatula_0",
    "can_12",
    "apple_7",
    "baguette_6",
    "coffee_cup_13",
    "banana_12",
    "knife_16",
    "cupcake_7",
    "potato_4",
    "cereal_13",
    "bread_7",
    "beer_6",
    "alcohol_0",
    "potato_8",
    "rolling_pin_2",
    "bottled_water_3",
    "bread_2",
    "kiwi_6",
    "bagged_food_0",
    "pan_3",
    "banana_14",
    "milk_0",
    "pan_15",
    "pear_13",
    "cereal_3",
    "chips_7",
    "alcohol_14",
    "garlic_2",
    "soap_dispenser_4",
    "mushroom_6",
    "cereal_10",
    "cake_8",
    "bagged_food_5",
    "beer_5",
    "donut_0",
    "bread_11",
    "spoon_9",
    "cucumber_3",
    "boxed_food_7",
    "spoon_12",
    "potato_16",
    "mango_19",
    "lemon_11",
    "chocolate_10",
    "spatula_11",
    "chocolate_4",
    "banana_1",
    "can_6",
    "croissant_2",
    "donut_4",
    "banana_4",
    "bagel_7",
    "corn_1",
    "mug_11",
    "chips_5",
    "corn_4",
    "potato_5",
    "rolling_pin_0",
    "bowl_5",
    "banana_8",
    "knife_8",
    "mug_1",
    "knife_4",
    "scissors_12",
    "mug_15",
    "teapot_0",
    "canned_food_18",
    "alcohol_9",
    "can_4",
    "croissant_8",
    "pan_6",
    "alcohol_4",
    "pear_0",
    "baguette_5",
    "bottled_water_10",
    "mango_18",
    "tomato_3",
    "broccoli_1",
    "donut_12",
    "orange_10",
    "bar_11",
    "ketchup_11",
    "bottled_water_17",
    "coffee_cup_2",
    "avocado_0",
    "candle_4",
    "bottled_drink_4",
    "knife_11",
    "water_bottle_16",
    "pear_11",
    "pan_17",
    "mug_0",
    "wine_4",
    "boxed_food_6",
    "banana_9",
    "onion_4",
    "tangerine_5",
    "chips_3",
    "chocolate_2",
    "teapot_6",
    "cereal_1",
    "can_16",
    "spoon_10",
    "spatula_3",
    "bread_16",
    "tray_2",
    "potato_6",
    "pan_24",
    "onion_10",
    "candle_9",
    "fish_2",
    "egg_9",
    "chocolate_5",
    "apple_11",
    "ladle_3",
    "kettle_20",
    "cereal_9",
    "bottled_water_4",
    "donut_5",
    "cake_6",
    "beer_2",
    "alcohol_1",
    "kettle_5",
    "canned_food_0",
    "rolling_pin_3",
    "pear_5",
    "lemon_1",
    "fish_9",
    "mug_4",
    "peach_4",
    "teapot_5",
    "boxed_food_4",
    "pan_22",
    "tray_10",
    "knife_7",
    "avocado_8",
    "coffee_cup_6",
    "jug_0",
    "teapot_11",
    "banana_2",
    "cutting_board_2",
    "cupcake_2",
    "mug_7",
    "apple_22",
    "pan_1",
    "cereal_0",
    "steak_6",
    "croissant_5",
    "bread_17",
    "orange_11",
    "water_bottle_15",
    "cutting_board_9",
    "squash_5",
    "potato_18",
    "cereal_6",
    "croissant_3",
    "bottled_water_16",
    "kettle_12",
    "squash_7",
    "lemon_5",
    "fork_8",
    "scissors_3",
    "apple_21",
    "kettle_11",
    "spoon_21",
    "knife_0",
    "banana_19",
    "bread_8",
    "bagel_4",
    "mushroom_14",
    "canned_food_15",
    "fish_10",
    "orange_4",
    "canned_food_11",
    "knife_25",
    "steak_10",
    "chips_11",
    "water_bottle_3",
    "cupcake_16",
    "water_bottle_2",
    "bowl_4",
    "milk_10",
    "knife_13",
    "cheese_3",
    "kettle_27",
    "kettle_22",
    "kiwi_5",
    "spray_1",
    "apple_19",
    "spoon_3",
    "lemon_6",
    "orange_9",
    "pan_11",
    "cupcake_20",
    "orange_5",
    "baguette_4",
    "chips_10",
    "egg_0",
    "cutting_board_15",
    "milk_1",
    "can_9",
    "apple_12",
    "fork_3",
    "sponge_1",
    "tray_4",
    "fish_8",
    "spatula_7",
    "apple_15",
    "lemon_17",
    "boxed_drink_1",
    "wine_10",
    "donut_3",
    "coffee_cup_9",
    "can_8",
    "tomato_5",
    "jug_2",
    "alcohol_11",
    "mango_11",
    "plate_7",
    "onion_6",
    "rolling_pin_4",
    "teapot_13",
    "eggplant_0",
    "chips_4",
    "lemon_15",
    "chips_9",
    "mushroom_2",
    "lemon_13",
    "carrot_6",
    "bagged_food_14",
    "coffee_cup_15",
    "cucumber_0",
    "mushroom_8",
    "potato_17",
    "jam_3",
    "milk_4",
    "cupcake_19",
    "pear_10",
    "canned_food_5",
    "cutting_board_1",
    "candle_6",
    "egg_8",
    "carrot_0",
    "bar_8",
    "knife_24",
    "mushroom_15",
    "boxed_drink_5",
    "apple_1",
    "bowl_8",
    "chocolate_14",
    "water_bottle_9",
    "banana_3",
    "kettle_28",
    "jug_6",
    "chocolate_9",
    "mango_14",
    "banana_7",
    "peach_6",
    "alcohol_10",
    "potato_10",
    "cereal_2",
    "ketchup_3",
    "kettle_14",
    "chips_13",
    "egg_10",
    "plate_2",
    "bagged_food_3",
    "onion_1",
    "mango_1",
    "pear_15",
    "bell_pepper_1",
    "can_18",
    "mug_16",
    "fish_5",
    "bread_3",
    "donut_11",
    "bottled_water_7",
    "pan_2",
    "shaker_3",
    "tray_5",
    "coffee_cup_1",
    "bar_soap_4",
    "shaker_1",
    "banana_11",
    "plate_4",
    "kettle_6",
    "mug_10",
    "pan_7",
    "pan_20",
    "water_bottle_0",
    "garlic_0",
    "cupcake_3",
    "croissant_6",
    "tangerine_6",
    "bottled_drink_5",
    "cheese_0",
    "bar_6",
    "candle_8",
    "candle_15",
    "boxed_drink_4",
    "spatula_8",
    "donut_10",
    "beer_19",
    "cutting_board_7",
    "banana_24",
    "pan_4",
    "bagged_food_9",
    "pan_14",
    "ketchup_10",
    "tray_0",
    "bar_soap_1",
    "candle_7",
    "sponge_5",
    "steak_11",
    "bar_soap_0",
    "onion_8",
    "chocolate_0",
    "rolling_pin_1",
    "cake_9",
    "bowl_10",
    "knife_10",
    "chips_16",
    "egg_4",
    "jam_5",
    "chips_14",
    "steak_9",
    "wine_5",
    "spoon_20",
    "mug_17",
    "boxed_food_11",
    "donut_2",
    "can_15",
    "apple_0",
    "cupcake_4",
    "apple_14",
    "carrot_7",
    "steak_8",
    "mango_13",
    "avocado_4",
    "condiment_3",
    "potato_19",
    "baguette_2",
    "teapot_10",
    "chips_6",
    "water_bottle_22",
    "soap_dispenser_0",
    "carrot_2",
    "boxed_food_10",
    "spoon_11",
    "banana_22",
    "hot_dog_2",
    "boxed_food_8",
    "egg_7",
    "waffle_3",
    "fork_10",
    "knife_22",
    "pear_2",
    "milk_7",
    "bagel_3",
    "tray_3",
    "knife_21",
    "cupcake_14",
    "bottled_water_8",
    "baguette_1",
    "tomato_0",
    "cupcake_11",
    "plate_20",
    "pan_5",
    "chips_15",
    "kiwi_4",
    "bowl_7",
    "mug_14",
    "kettle_21",
    "bagel_1",
    "cup_4",
    "apple_18",
    "apple_5",
    "fork_7",
    "mushroom_1",
    "cheese_5",
    "kiwi_2",
    "soap_dispenser_1",
    "ketchup_1",
    "beer_13",
    "bagel_0",
    "mango_10",
    "spatula_2",
    "knife_26",
    "orange_8",
    "avocado_6",
    "orange_2",
    "croissant_10",
    "knife_23",
    "bottled_water_14",
    "steak_5",
    "mug_5",
    "cereal_12",
    "spoon_16",
    "bar_0",
    "mug_2",
    "bread_10",
    "apple_17",
    "mug_9",
    "wine_13",
    "spoon_7",
    "bread_9",
    "donut_6",
    "hot_dog_6",
    "apple_2",
    "wine_9",
    "orange_12",
    "cupcake_5",
    "fish_4",
    "chips_0",
    "coffee_cup_11",
    "banana_10",
    "sponge_0",
    "fish_6",
    "jam_0",
    "apple_3",
    "pan_0",
    "apple_10",
    "potato_0",
    "bottled_water_1",
    "squash_13",
    "coffee_cup_10",
    "beer_11",
    "potato_11",
    "lime_1",
    "pear_19",
    "spoon_15",
    "candle_5",
    "tray_12",
    "bottled_water_11",
    "pear_17",
    "egg_6",
    "canned_food_3",
    "steak_3",
    "mushroom_5",
    "bowl_14",
    "coffee_cup_14",
    "tangerine_0",
    "bell_pepper_3",
    "ladle_1",
    "soap_dispenser_6",
    "plate_9",
    "scissors_9",
    "fork_9",
    "kettle_16",
    "eggplant_3",
    "fork_2",
    "corn_3",
    "boxed_food_2",
    "mug_3",
    "pan_13",
    "cutting_board_0",
    "avocado_1",
    "water_bottle_4",
    "mango_7",
    "knife_20",
    "lemon_12",
    "kettle_18",
    "cake_5",
    "teapot_14",
    "kettle_2",
    "tomato_2",
    "beer_18",
    "orange_1",
    "plate_19",
    "hot_dog_5",
    "egg_1",
    "banana_15",
    "mango_8",
    "onion_3",
    "steak_15",
    "mango_15",
    "tomato_7",
    "spoon_13",
    "coffee_cup_20",
    "eggplant_2",
    "onion_2",
    "coffee_cup_4",
    "beer_1",
    "spoon_4",
    "squash_1",
    "bagel_5",
    "bowl_15",
    "spatula_10",
    "sponge_3",
    "boxed_food_0",
    "bar_9",
    "kettle_7",
    "jam_1",
    "potato_14",
    "knife_12",
    "pear_3",
    "cutting_board_11",
    "bagged_food_4",
    "spoon_2",
    "plate_5",
    "alcohol_2",
    "beer_15",
    "coffee_cup_7",
    "bowl_0",
    "scissors_11",
    "pear_1",
    "bread_20",
    "canned_food_17",
    "alcohol_12",
    "bread_5",
    "shaker_2",
    "candle_0",
    "garlic_4",
    "tray_7",
    "potato_9",
    "canned_food_13",
    "bowl_2",
    "beer_10",
    "cucumber_2",
    "steak_14",
    "cupcake_15",
    "onion_7",
    "tomato_1",
    "donut_15",
    "boxed_food_9",
    "spoon_23",
    "banana_20",
    "yogurt_2",
    "mug_12",
    "egg_11",
    "lemon_0",
    "kettle_10",
    "donut_8",
    "eggplant_1",
    "mug_13",
    "cereal_4",
    "alcohol_5",
    "knife_9",
    "tangerine_2",
    "bar_7",
    "ladle_4",
    "canned_food_4",
    "beer_3",
    "teapot_3",
    "cake_7",
    "soap_dispenser_3",
    "lemon_7",
    "spoon_0",
    "ketchup_4",
    "candle_2",
    "spoon_19",
    "cup_2",
    "ketchup_7",
    "kiwi_1",
    "apple_6",
    "jug_5",
    "apple_20",
    "bagged_food_7",
    "tangerine_1",
    "fish_11",
    "rolling_pin_5",
    "pan_10",
    "cheese_7",
    "lemon_3",
    "knife_18",
    "candle_14",
    "canned_food_14",
    "condiment_2",
    "lemon_14",
    "plate_8",
    "wine_8",
    "spoon_1",
    "bread_0",
    "canned_food_1",
    "kettle_26",
    "cucumber_4",
    "cupcake_12",
    "canned_food_19",
    "bagged_food_11",
    "scissors_1",
    "mushroom_9"
]